/*

    File: readpart.c

    Copyright (C) 1998-2004 Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
 
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <ctype.h>      /* toupper, tolower */
#include "types.h"
#include "common.h"
#include "readpart.h"
#include "testdisk.h"
#include "fnctdsk.h"
#include "analyse.h"
#include "lang.h"
#include "intrf.h"
typedef struct {
  unsigned char info[128];   /* Informative text string */
  unsigned char spare0[14];
  struct sun_info {
    unsigned char spare1;
    unsigned char id;
    unsigned char spare2;
    unsigned char flags;
  } infos[8];
  unsigned char spare1[246]; /* Boot information etc. */
  unsigned short rspeed;     /* Disk rotational speed */
  unsigned short pcylcount;  /* Physical cylinder count */
  unsigned short sparecyl;   /* extra sects per cylinder */
  unsigned char spare2[4];   /* More magic... */
  unsigned short ilfact;     /* Interleave factor */
  unsigned short ncyl;       /* Data cylinder count */
  unsigned short nacyl;      /* Alt. cylinder count */
  unsigned short ntrks;      /* Tracks per cylinder */
  unsigned short nsect;      /* Sectors per track */
  unsigned char spare3[4];   /* Even more magic... */
  struct sun_partition {
    uint32_t start_cylinder;
    uint32_t num_sectors;
  } partitions[8];
  unsigned short magic;      /* Magic number */
  unsigned short csum;       /* Label xor'd checksum */
} sun_partition;

#define SUNOS_SWAP 3
#define WHOLE_DISK 5
#define SUN_LABEL_MAGIC          0xDABE

static t_list_part *get_ext_data(t_param_disk *disk_car, t_list_part *list_part, const int debug, const int saveheader);
static void test_MBR_data(t_list_part *list_part);
static int test_MBR_over(t_param_disk *disk_car,t_list_part *list_part);
static int write_mbr(t_param_disk *disk_car, const t_list_part *list_part, const int ro, const int debug);
static int write_all_log(t_param_disk *disk_car, const t_list_part *list_part, const int ro , const int debug, const int align);
static int diff(const t_param_disk *disk_car,const unsigned char buffer[DEFAULT_SECTOR_SIZE], const unsigned char buffer_org[DEFAULT_SECTOR_SIZE]);

static t_list_part *read_part_i386(t_param_disk *disk_car, const int debug, const int saveheader);
static int write_part_i386(t_param_disk *disk_car, const t_list_part *list_part, const int ro , const int debug, const int align);

static t_list_part *read_part_sun(t_param_disk *disk_car, const int debug, const int saveheader);
static int write_part_sun(t_param_disk *disk_car, const t_list_part *list_part, const int ro , const int debug, const int align);

t_arch_fnct arch_i386= {
  .part_sys=PART_SYS_i386,
  .read_part=read_part_i386,
  .write_part=write_part_i386
};


t_arch_fnct arch_sun=
{
  .part_sys=PART_SYS_SUN,
  .read_part=read_part_sun,
  .write_part=write_part_sun
};


int get_geometry_from_mbr(const unsigned char *buffer, const int debug, t_CHS *geometry)
{
  unsigned int i;
  geometry->cylinder=0;
  geometry->head=0;
  geometry->sector=0;
  if((buffer[0x1FE]!=(unsigned char)0x55)||(buffer[0x1FF]!=(unsigned char)0xAA))
  {
    return 1;
  }
  for(i=0;i<4;i++)
  {
    const struct partition_dos *p=pt_offset_const(buffer,i);
    if(p->sys_ind!=0)
    {
      if(geometry->cylinder<e_cyl(p))
	geometry->cylinder=e_cyl(p);
      if(geometry->head<e_head(p))
	geometry->head=e_head(p);
      if(geometry->sector<e_sect(p))
	geometry->sector=e_sect(p);
    }
  }
  if(geometry->sector>0)
  {
    ecrit_rapport("Geometry from MBR: head=%u sector=%u\n",geometry->head+1,geometry->sector);
  }
  return 0;
}

t_list_part *read_part_sun(t_param_disk *disk_car, const int debug, const int saveheader)
{
  unsigned char buffer[DEFAULT_SECTOR_SIZE];
  sun_partition *sunlabel=(sun_partition*)&buffer;
  t_list_part *new_list_part=NULL;
  unsigned int i;
  aff_buffer(BUFFER_RESET,"Q");
  if(read_MBR(disk_car,&buffer)!=0)
    return NULL;
  if (sunlabel->magic != SUN_LABEL_MAGIC)
  {
    aff_buffer(BUFFER_ADD,"Bad SUN partition\n");
    return NULL;
  }
  for(i=0;i<8;i++)
  {
     if (sunlabel->partitions[i].num_sectors > 0
	 && sunlabel->infos[i].id > 0
	 && sunlabel->infos[i].id != WHOLE_DISK)
     {
       t_diskext *new_partition=partition_new();
       partition_reset(new_partition);
       new_partition->order=i;
       new_partition->part_type=sunlabel->infos[i].id;
       new_partition->part_offset=sunlabel->partitions[i].start_cylinder*sunlabel->ntrks*sunlabel->nsect*disk_car->sector_size;
       new_partition->part_size=(uint64_t)sunlabel->partitions[i].num_sectors*disk_car->sector_size;
       new_partition->status=STATUS_PRIM;
       aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
       new_list_part=insert_new_partition(new_list_part,new_partition);
     }
  }
  return new_list_part;
}

static int write_part_sun(t_param_disk *disk_car, const t_list_part *list_part, const int ro, const int debug, const int align)
{
  return 0;
}

static t_list_part *read_part_i386(t_param_disk *disk_car, const int debug, const int saveheader)
{
  unsigned char buffer[disk_car->sector_size];
  t_list_part *new_list_part=NULL;
  unsigned int i;
  int res=0;
  t_CHS geometry;
  aff_buffer(BUFFER_RESET,"Q");
  if(read_MBR(disk_car,&buffer)!=0)
    return NULL;
  if(get_geometry_from_mbr(buffer,debug,&geometry)!=0)
  {
    aff_buffer(BUFFER_ADD,msg_TBL_NMARK);
    return NULL;
  }
  for(i=0;i<4;i++)
  {
    const struct partition_dos *p=pt_offset(buffer,i);
    status_type_t status=STATUS_PRIM;
    switch(p->sys_ind)
    {
      case P_NO_OS:
	break;
      case P_EXTENDX:
      case P_EXTENDED:
      case P_LINUXEXTENDX:
	status=STATUS_EXT;
	/* don't put a break */
      default:
	{
	  t_diskext *new_partition=partition_new();
	  i386_entry2partition(disk_car, (uint64_t)0, new_partition, p, status,i+1,debug,saveheader);
	  if(debug>1)
	    aff_entry_rapport(disk_car,p);
	  aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
	  if(new_partition->errcode!=BAD_NOERR)
	  {
	    aff_buffer(BUFFER_ADD,"%s\n",errmsg_i386_entry2partition(new_partition->errcode));
	    res|=1;
	  }
	  new_list_part=insert_new_partition(new_list_part,new_partition);
	}
	break;
    }
  }
/*  if(res==0) */
  {
    test_MBR_data(new_list_part);
    if(test_MBR_over(disk_car,new_list_part)==0)
      new_list_part=get_ext_data(disk_car,new_list_part,debug,saveheader);
  }
  get_geometry_from_list_part(disk_car, new_list_part, debug);
  return new_list_part;
}

static void test_MBR_data(t_list_part *list_part)
{
  t_list_part *element;
  unsigned int nb_dos=0, nb_hidden=0, nb_mb=0, nb_ext=0, nb_boot=0;
  wdoprintf(stdscr,msg_PART_HEADER_LONG);
  for(element=list_part;element!=NULL;element=element->next)
  {
    const t_diskext *partition=element->part;
    switch(partition->status)
    {
      case STATUS_PRIM_BOOT:
	nb_boot++;
      case STATUS_PRIM:
	switch(partition->part_type)
	{
	  case P_12FAT:
	  case P_16FAT:
	  case P_16FATBD:
	    nb_dos++;
	    break;
	  case P_16FATBDH:
	  case P_16FATH:
	  case P_NTFSH:
	    nb_hidden++;
	    break;
	  case P_OS2MB:
	    nb_mb++;
	    break;
	}
	break;
      case STATUS_EXT:
	nb_ext++;
	break;
      default:
	ecrit_rapport("test_MBR_data: severe error\n");
	break;
    }
  }
  if(nb_dos>1)
    aff_buffer(BUFFER_ADD,msg_ONLY_ONE_DOS);
  if(nb_ext>1)
    aff_buffer(BUFFER_ADD,msg_ONLY_ONE_EXT);
  /* S'il y a des partitions caches, il faut un MB */
  /* Obsolete
  if(nb_hidden>0 && nb_mb==0)
    aff_buffer(BUFFER_ADD,msg_NO_OS2MB);
    */
  /* Nombre de partition bootable */
  if(nb_boot==0)
    aff_buffer(BUFFER_ADD,msg_NO_BOOTABLE);
  else
    if(nb_boot>1)
      aff_buffer(BUFFER_ADD,msg_ONLY1MUSTBOOT);
}

static t_list_part *get_ext_data(t_param_disk *disk_car, t_list_part *list_part, const int debug, const int saveheader)
{
  t_list_part *element;
  int res=0;
  t_diskext *partition_main_ext=NULL;
  for(element=list_part;element!=NULL;element=element->next)
  {
    if(element->part->status==STATUS_EXT)
      partition_main_ext=element->part;
  }
  if(partition_main_ext!=NULL)
  {
    t_diskext *partition_ext=partition_main_ext;
    t_diskext *partition_next_ext=NULL;
    unsigned int i;
    unsigned int order=5;
    do
    {
      unsigned char buffer[DEFAULT_SECTOR_SIZE];
      int nb_hidden=0, nb_mb=0, nb_part=0, nb_ext=0, nb_boot=0;
      if(disk_car->read(disk_car,sizeof(buffer), &buffer, partition_ext->part_offset)!=0)
	return list_part;
      if((buffer[0x1FE]!=(unsigned char)0x55)||(buffer[0x1FF]!=(unsigned char)0xAA))
      {
	aff_buffer(BUFFER_ADD,"\ntest_logical: " msg_TBL2_NMARK);
	return list_part;
      }
      for(i=0;i<4;i++)
      {
	const struct partition_dos *p=pt_offset(buffer,i);
	if(p->boot_ind==(unsigned char)0x80)
	  nb_boot++;
	switch(p->sys_ind)
	{
	  case P_16FATBDH:
	  case P_16FATH:
	  case P_NTFSH:
	    nb_hidden++;
	    break;
	  case P_OS2MB:
	    nb_mb++;
	    break;
	  case P_EXTENDX:
	  case P_EXTENDED:
	  case P_LINUXEXTENDX:
	    nb_ext++;
	    break;
	  case P_NO_OS:
	    break;
	  default:
	    nb_part++;
	}
      }
      if(nb_hidden>0)
	aff_buffer(BUFFER_ADD,"Partition musn't be hidden\n");
      if(nb_mb>0)
	aff_buffer(BUFFER_ADD,"Multiboot must be a primary partition, not a logical\n");
      if(nb_ext>1)
	aff_buffer(BUFFER_ADD,"A logical partition musn't have more than one link to another logical partition\n");
      if(nb_part>1)
	aff_buffer(BUFFER_ADD,"A logical partition must contain only one partition\n");
      if(nb_boot>0)
	aff_buffer(BUFFER_ADD,"Logical partition musn't be bootable\n");
      partition_next_ext=NULL;
      for(i=0;i<4;i++)
      {
	const struct partition_dos *p=pt_offset(buffer,i);
	if(p->sys_ind!=0)
	{
	  t_diskext *new_partition=partition_new();
	  new_partition->order=order;
	  if(debug>1)
	    aff_entry_rapport(disk_car,p);
	  if(is_extended(p->sys_ind))
	  {
	    partition_next_ext=new_partition;
	    i386_entry2partition(disk_car, partition_main_ext->part_offset, new_partition, p, STATUS_EXT_IN_EXT,order,debug,saveheader);
	    aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
	    if(new_partition->errcode!=BAD_NOERR)
	    {
	      aff_buffer(BUFFER_ADD,"%s\n",errmsg_i386_entry2partition(new_partition->errcode));
	      res=1;
	    }
	    {
	      if((new_partition->part_offset<=partition_main_ext->part_offset) ||
		  (new_partition->part_offset+new_partition->part_size-1 > partition_main_ext->part_offset+partition_main_ext->part_size-1))
	      {	/* Must be IN partition_main_ext */
		res=1;
		aff_buffer(BUFFER_ADD,"Must be in extended partition\n");
		aff_part_buffer(AFF_PART_ORDER,disk_car,partition_main_ext);
		aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
	      }
	      else
	      {
		for(element=list_part;element!=NULL;element=element->next)
		{
		  t_diskext *partition=element->part;
		  if(partition->status==STATUS_EXT_IN_EXT)
		  {
		    if(((partition->part_offset>=new_partition->part_offset) && (partition->part_offset<=new_partition->part_offset+new_partition->part_size-1)) ||
			((partition->part_offset+partition->part_size-1>=new_partition->part_offset) && (partition->part_offset+partition->part_size-1<=new_partition->part_offset+partition->part_size-1)))
		    { /* New Partition start or end musn't been in partition */
		      res=1;
		      aff_buffer(BUFFER_ADD, "Logical partition must be in its own extended partition\n");
		      aff_part_buffer(AFF_PART_ORDER,disk_car,partition);
		      aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
		    }
		  }
		}
	      }
	    }
	  }
	  else
	  {
	    i386_entry2partition(disk_car,partition_ext->part_offset, new_partition, p, STATUS_LOG,order,debug,saveheader);
	    order++;
	    if(debug>1)
	      aff_entry_rapport(disk_car,p);
	    aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
	    if(new_partition->errcode!=BAD_NOERR)
	    {
	      aff_buffer(BUFFER_ADD,"%s\n",errmsg_i386_entry2partition(new_partition->errcode));
	      res=1;
	    }
	    {
	      if((new_partition->part_offset<=partition_main_ext->part_offset) ||
		  (new_partition->part_offset+new_partition->part_size-1 > partition_main_ext->part_offset+partition_main_ext->part_size-1))
	      {	/* Must be IN partition_main_ext */
		res=1;
		aff_buffer(BUFFER_ADD, msg_SAME_SPACE);
		aff_part_buffer(AFF_PART_ORDER,disk_car,partition_main_ext);
		aff_part_buffer(AFF_PART_ORDER,disk_car,new_partition);
	      }
	    }
	  }
	  list_part=insert_new_partition(list_part,new_partition);
	}
      }
      partition_ext=partition_next_ext;
    } while ((partition_ext!=NULL) && (order<30));
  }
  return list_part;
}

static int test_MBR_over(t_param_disk *disk_car,t_list_part *list_part)
{/* Tester le chevauchement des partitions */
  int res=0;
  t_list_part *element;
  for(element=list_part;element!=NULL;element=element->next)
    if((element->next!=NULL) && (element->part->part_offset + element->part->part_size - 1 >= element->next->part->part_offset))
    {
      res=1;
      aff_buffer(BUFFER_ADD, msg_SAME_SPACE);
      aff_part_buffer(AFF_PART_ORDER,disk_car,element->part);
      aff_part_buffer(AFF_PART_ORDER,disk_car,element->next->part);
    }
  return res;
}

int write_part_i386(t_param_disk *disk_car, const t_list_part *list_part, const int ro, const int debug, const int align)
{
  write_mbr(disk_car,list_part,ro,debug);
  return write_all_log(disk_car,list_part,ro,debug,align);
}

static int write_mbr(t_param_disk *disk_car, const t_list_part *list_part, const int ro, const int debug)
{
  const t_list_part *element;
  unsigned char buffer[disk_car->sector_size];
  unsigned char buffer_org[disk_car->sector_size];
  if(debug>0)
  {
    ecrit_rapport("\nwrite_mbr: starting...\n");
  }
  if(read_MBR(disk_car,&buffer_org))
  {
    memset((void *)&buffer_org,0,DEFAULT_SECTOR_SIZE);
  }
  memset((void *)&buffer,0,DEFAULT_SECTOR_SIZE);
  if((buffer_org[0x1FE]==0x55) && (buffer_org[0x1FF]==0xAA))
  {
    memcpy(buffer,buffer_org,TAB_PART);
    buffer[0x1FE]=0x55;
    buffer[0x1FF]=0xAA;
  } else {
    write_MBR_code_aux(buffer);
  }
  for(element=list_part;element!=NULL;element=element->next)
  {
    switch(element->part->status)
    {
      case STATUS_PRIM:
      case STATUS_PRIM_BOOT:
      case STATUS_EXT:
	if((element->part->order>=1) && (element->part->order<=4))
	{
	  partition2entry(disk_car,(uint64_t)0,element->part, 
	      pt_offset(buffer,element->part->order-1));
	}
	break;
      case STATUS_LOG:
	break;
      default:
	ecrit_rapport("write_mbr: severe error\n");
	break;
    }
  }
  if(debug>1)
  {
    int i;
    for(i=0;i<4;i++)
    {
      const struct partition_dos *p=pt_offset(buffer,i);
      aff_entry_rapport(disk_car,p);
    }
    diff(disk_car,buffer,buffer_org);
  }
  if(ro==0)
    return write_MBR(disk_car,&buffer);
  return 0;
}

static int write_all_log(t_param_disk *disk_car, const t_list_part *list_part, const int ro, const int debug, const int align)
{
  const t_list_part *element;
  const t_list_part *pos_ext=NULL;
  uint64_t current_pos;
  t_diskext *bloc_nextext;
  if(debug)
    ecrit_rapport("write_all_log: starting...\n");
  for(element=list_part;element!=NULL;element=element->next)
  {
    if(element->part->status==STATUS_EXT)
    {
      if(is_extended(element->part->part_type))
      {
	if(pos_ext!=NULL)
	  ecrit_rapport("write_all_log: pos_ext already defined\n");
	pos_ext=element;
      }
      else
      {
	ecrit_rapport("write_all_log: STATUS_EXT with bad part_type\n");
      }
    }
  }
  if(pos_ext==NULL)
  {
    wdoprintf(stdscr,msg_NO_EXT_PART);
    return 1;
  }
  current_pos=pos_ext->part->part_offset;
  bloc_nextext=(t_diskext *)MALLOC(sizeof(*bloc_nextext));
  bloc_nextext->part_type=P_EXTENDED;       /* Never P_EXTENDX */
  if(pos_ext->next==NULL || (pos_ext->next->part->status!=STATUS_LOG))
  {
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    unsigned char buffer_org[DEFAULT_SECTOR_SIZE];
    if(debug>0)
    {
      ecrit_rapport("write_all_log: CHS: %u/%u/%u,lba=%lu\n", offset2cylinder(disk_car,current_pos), offset2head(disk_car,current_pos), offset2sector(disk_car,current_pos),(long unsigned)(current_pos/disk_car->sector_size));
    }
    if(disk_car->read(disk_car,sizeof(buffer_org), &buffer_org, current_pos))
    {
      memset(buffer_org,0,DEFAULT_SECTOR_SIZE);
    }
    memset(buffer,0,DEFAULT_SECTOR_SIZE);
    memcpy(buffer,buffer_org,TAB_PART);
    buffer[0x1FE]=0x55;
    buffer[0x1FF]=0xAA;
    if(ro)
    {
      if(debug>1)
	diff(disk_car,buffer,buffer_org);
    }
    else
    {
      if(disk_car->write(disk_car,sizeof(buffer), &buffer, current_pos))
      {
	ecrit_rapport(msg_PART_WR_ERR);
	display_message(msg_PART_WR_ERR);
      }
    }
  }
  else
  {
    for(element=pos_ext->next;(element!=NULL) && (element->part->status==STATUS_LOG);element=element->next)
    {
      int j;
      unsigned char buffer[DEFAULT_SECTOR_SIZE];
      unsigned char buffer_org[DEFAULT_SECTOR_SIZE];
      if(debug>0)
      {
	ecrit_rapport("write_all_log: CHS: %u/%u/%u,lba=%lu\n", offset2cylinder(disk_car,current_pos), offset2head(disk_car,current_pos), offset2sector(disk_car,current_pos),(long unsigned)(current_pos/disk_car->sector_size));
      }
      if(disk_car->read(disk_car,sizeof(buffer_org), &buffer_org, current_pos))
      {
	memset(buffer_org,0,DEFAULT_SECTOR_SIZE);
      }
      memset(buffer,0,DEFAULT_SECTOR_SIZE);
      memcpy(buffer,buffer_org,TAB_PART);
      buffer[0x1FE]=0x55;
      buffer[0x1FF]=0xAA;
      partition2entry(disk_car,current_pos,element->part, pt_offset(buffer,0));
      if(element->next!=NULL && (element->next->part->status==STATUS_LOG))
      { /* Construit le pointeur vers la prochaine partition logique */
	t_CHS nextext_start;
	bloc_nextext->part_offset=element->next->part->part_offset-disk_car->sector_size;
	offset2CHS(disk_car,bloc_nextext->part_offset,&nextext_start);
	nextext_start.sector=1;
	if(nextext_start.head>0)
	  nextext_start.head--;
	else
	{
	  nextext_start.head=disk_car->CHS.head;
	  nextext_start.cylinder--;
	}
	if(CHS2offset(disk_car,&nextext_start)<=element->part->part_offset+element->part->part_size+1)
	{
	  offset2CHS(disk_car,bloc_nextext->part_offset,&nextext_start);
	  nextext_start.sector=1;
	  if(CHS2offset(disk_car,&nextext_start)<=element->part->part_offset+element->part->part_size+1)
	  {
	    offset2CHS(disk_car,bloc_nextext->part_offset,&nextext_start);
	  }
	}
	bloc_nextext->part_offset=CHS2offset(disk_car,&nextext_start);
	/*      ecrit_rapport("table[i]->next=%p table[i+1]=%p\n",table[i]->next,table[i+1]); */
	bloc_nextext->part_size=(uint64_t)element->next->part->part_offset+element->next->part->part_size-bloc_nextext->part_offset;
	partition2entry(disk_car,pos_ext->part->part_offset,bloc_nextext, pt_offset(buffer,1));
      }
      if(ro)
      {
	if(debug>1)
	{
	  for(j=0;j<4;j++)
	  {
	    const struct partition_dos *p=pt_offset(buffer,j);
	    if(p->sys_ind!=0)
	      aff_entry_rapport(disk_car,p);
	  }
	  diff(disk_car,buffer,buffer_org);
	}
      }
      else
      {
	if(disk_car->write(disk_car,sizeof(buffer), &buffer, current_pos))
	{
	  ecrit_rapport(msg_PART_WR_ERR);
	  display_message(msg_PART_WR_ERR);
	}
      }
      current_pos=bloc_nextext->part_offset;
    }
  }
  FREE(bloc_nextext);
  return 0;
}

static int diff(const t_param_disk *disk_car, const unsigned char buffer[DEFAULT_SECTOR_SIZE], const unsigned char buffer_org[DEFAULT_SECTOR_SIZE])
{
  int j;
  if(memcmp(buffer,buffer_org,DEFAULT_SECTOR_SIZE))
  {
    ecrit_rapport("\nSectors are different.\n");
    ecrit_rapport("buffer_org\n");
    for(j=0;j<4;j++)
    {
      const struct partition_dos *p=pt_offset_const(buffer_org,j);
      if(p->sys_ind!=0)
	aff_entry_rapport(disk_car,p);
    }
    ecrit_rapport("buffer\n");
    for(j=0;j<4;j++)
    {
      const struct partition_dos *p=pt_offset_const(buffer,j);
      if(p->sys_ind!=0)
	aff_entry_rapport(disk_car,p);
    }
    for(j=0;j<DEFAULT_SECTOR_SIZE;j++)
      if(buffer_org[j]!=buffer[j])
	ecrit_rapport("%02X %02X %02X\n", j, buffer_org[j], buffer[j]);
    ecrit_rapport("\n");
  }
  return 0;
}

